#                       Copyright 2021 Saad Mairaj

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import torch
from transformers import AutoTokenizer, AutoModelForSequenceClassification
from boolean_question.utils import get_model, set_seed, device
from collections import OrderedDict


class BoolQ:
    """Boolean question model

    Predict answer as "True" or "False" by providing text 
    passage and a question related to the text passage 
    to the model. 

    Also, get the confidence for both "True" and "False"
    answer using `prediction_details` method. Else use just 
    `predict` method to get the most confident answer."""

    def __init__(self, model: str = None, confidence: float = None, seed: int = 26):
        """Initialise the BoolQ model to predict question from text passage.

        Args:
            model (str): Give path to the transformer model. By default \
                              is None.
            confidence (float): Give confidence level between (0.1 - 1.0) of \
                                the answer.By default is None, means will \
                                return the answer with maximum confidence.
            seed (int): Change seed for reproducibility. By default is 26.        
        """
        # Use a GPU if you have one available (Runtime -> Change runtime type -> GPU)
        model = model or get_model()

        # Set seeds for reproducibility
        set_seed(seed)

        self.tokenizer = AutoTokenizer.from_pretrained(model)
        self.model = AutoModelForSequenceClassification.from_pretrained(model)
        self.model.to(device=device())

        self._conf = confidence

    def predict(self, passage: str, question: str):
        """Predict if the given question is true or false from the passage.

        Args:
            passage (str): Give an input text passage.
            quesion (str): Give a question related to the given text passage.

        Returns:
            Boolean: Either True for correct question or False for wrong question."""

        self.passage = passage
        self.question = question

        sequence = self.tokenizer.encode_plus(
            question, passage, return_tensors="pt")['input_ids'].to(device())

        logits = self.model(sequence)[0]
        probabilities = torch.softmax(logits, dim=1).detach().cpu().tolist()[0]
        self.probability = OrderedDict()
        self.probability[True] = round(probabilities[1], 2)
        self.probability[False] = round(probabilities[0], 2)

        self.answer = [k for k, v in self.probability.items(
        ) if v == max(self.probability.values())][0]
        if self._conf is None or self._conf <= self.answer:
            return self.answer
        return None

    def prediction_details(self):
        """Get complete prediction details"""
        if getattr(self, 'probability', None) is None:
            raise RuntimeError(
                "No answer prediction found. Please run predict first")
        data = {}
        data['passage'] = self.passage
        data['question'] = self.question
        data['confidence'] = self._conf
        data['true confidence'], data['false confidence'] = list(
            self.probability.values())
        data['answer'] = self.answer
        return data



#                       Copyright 2021 Saad Mairaj

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

import os
import torch
import random
import numpy as np
import gdown
import shutil


def device():
    """Returns CUDA if GPU is present else returns CPU"""
    return torch.device("cuda" if torch.cuda.is_available() else "cpu")


def set_seed(seed):
    """Completely reproducible results are not guaranteed across 
    PyTorch releases, individual commits, or different platforms. 
    Furthermore, results may not be reproducible between CPU 
    and GPU executions, even when using identical seeds.

    However, there are some steps you can take to limit the 
    number of sources of nondeterministic behavior for a 
    specific platform, device, and PyTorch release. 
    First, you can control sources of randomness that can 
    cause multiple executions of your application to behave 
    differently. Second, you can configure PyTorch to avoid 
    using nondeterministic algorithms for some operations, 
    so that multiple calls to those operations, given the
    same inputs, will produce the same result.

    Sets seed for random, np.random and for torch.manual_seed. 
    If GPU is present then it torch.manual_seed_all is also set 
    to the same value"""
    random.seed(26)
    np.random.seed(26)
    torch.manual_seed(26)
    if torch.cuda.is_available():
        torch.cuda.manual_seed_all(seed)


def get_model(overwite=False):
    """Returns pretained model path or if not exists 
    downloads te model else raise FileNotFoundError."""
    relative_path = os.path.join(os.path.abspath(os.path.dirname(
        os.path.dirname(os.path.dirname(__file__)))), "model")
    if not os.path.exists(relative_path) or overwite:
        file_url = "https://drive.google.com/u/3/uc?id=1nJh6vjVBO8P0mT0IYAxNYqrPo7Uj2j94"
        filename = relative_path + '.zip'
        if not os.path.exists(filename) or overwite:
            filename = gdown.download(file_url, filename, use_cookies=True)
            if filename is None:
                raise RuntimeError("Unable to download the model")
        gdown.extractall(filename)
        temp_dir = os.path.join(os.path.dirname(relative_path), "__MACOSX")
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)
    return os.path.abspath(relative_path)



from setuptools import setup, find_packages

__author__ = "Saad Mairaj"
__version__ = "0.0.1"


def get_long_description(path):
    """Opens and fetches text of long descrition file."""
    with open(path, 'r') as f:
        return f.read()


attrs = dict(
    name='boolean_question',
    version=__version__,
    packages=find_packages(),
    long_description=get_long_description('README.md'),
    description='Boolean question-answer prediction with AI',
    long_description_content_type='text/markdown',
    author=__author__,
    author_email='Saadmairaj@yahoo.in',
    url='https://github.com/Saadmairaj/boolean-question',
    license='Apache',
    python_requires='>=3',
    install_requires=[
        'gdown',
        'torch',
        'torchvision',
        'transformers',
        'pandas',
        'numpy',
    ],
    keywords=[
        'question-answer',
        'nlp',
        'pytorch',
        'AI',
        'artificial-intelligence',
        'deep-learning',
        'natural-language-processing'
    ],
    classifiers=[
        'Development Status :: 5 - Production/Stable',
        'Intended Audience :: Developers',
        'Intended Audience :: Education',
        "Intended Audience :: Science/Research",
        'Operating System :: OS Independent',
        'Programming Language :: Python :: 3',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
        'License :: OSI Approved :: Apache Software License',
    ],
    project_urls={
        'Bug Reports': 'https://github.com/Saadmairaj/boolean-question/issues',
        'Source': 'https://github.com/Saadmairaj/boolean-question',
        'Ask Question': "https://github.com/Saadmairaj/boolean-question/discussions",
    },
    include_package_data=True,
)

setup(**attrs)



#                       Copyright 2021 Saad Mairaj

#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at

#        http://www.apache.org/licenses/LICENSE-2.0

#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.

"""
Get accurate answer prediction for True or False question using this
python pytorch model. The model takes a passage and question as 
input and returns either "True" or "False" as predicted answer.
"""

__author__ = "Saad Mairaj"
__version__ = "0.0.1"

from boolean_question.boolq import BoolQ
